@@ -24,12 +24,12 @@ static GKeyFile *state;
 static gchar *state_filename;
 static GtkWindow *login_window, *panel_window;
 static GtkLabel *message_label, *prompt_label;
-static GtkTreeView *user_view;
 static GtkWidget *login_box, *prompt_box;
 static GtkEntry *prompt_entry;
+static GtkComboBox *user_combo;
 static GtkComboBox *session_combo;
 static GtkComboBox *language_combo;
-static gchar *default_font_name, *default_theme_name;
+static gchar *default_font_name, *default_theme_name, *default_icon_theme_name;
 static gboolean cancelling = FALSE, prompted = FALSE;
 
 static gchar *
@@ -52,6 +52,7 @@ set_session (const gchar *session)
     GtkTreeModel *model = gtk_combo_box_get_model (session_combo);
     GtkTreeIter iter;
     const gchar *default_session;
+    gchar *last_session;
 
     if (session && gtk_tree_model_get_iter_first (model, &iter))
     {
@@ -70,7 +71,15 @@ set_session (const gchar *session)
         } while (gtk_tree_model_iter_next (model, &iter));
     }
 
-    /* If failed to find this session, then try the default */
+    /* If failed to find this session, then try the previous, then the default */
+    last_session = g_key_file_get_value (state, "greeter", "last-session", NULL);
+    if (last_session && g_strcmp0 (session, last_session) != 0)
+    {
+        set_session (last_session);
+        g_free (last_session);
+        return;
+    }
+    g_free (last_session);
     default_session = lightdm_greeter_get_default_session_hint (greeter);
     if (default_session && g_strcmp0 (session, default_session) != 0)
     {
@@ -204,7 +213,7 @@ cancel_authentication (void)
     else
     {
         gtk_widget_hide (login_box);
-        gtk_widget_grab_focus (GTK_WIDGET (user_view));
+        gtk_widget_grab_focus (GTK_WIDGET (user_combo));
     }
 }
 
@@ -213,6 +222,9 @@ start_session (void)
 {
     gchar *language;
     gchar *session;
+    gchar *data;
+    gsize data_length;
+    GError *error = NULL;
 
     language = get_language ();
     if (language)
@@ -220,6 +232,23 @@ start_session (void)
     g_free (language);
 
     session = get_session ();
+
+    /* Remember last choice */
+    g_key_file_set_value (state, "greeter", "last-session", session);
+
+    data = g_key_file_to_data (state, &data_length, &error);
+    if (error)
+        g_warning ("Failed to save state file: %s", error->message);
+    g_clear_error (&error);
+    if (data)
+    {
+        g_file_set_contents (state_filename, data, data_length, &error);
+        if (error)
+            g_warning ("Failed to save state file: %s", error->message);
+        g_clear_error (&error);
+    }
+    g_free (data);
+
     if (!lightdm_greeter_start_session_sync (greeter, session, NULL))
     {
         set_message_label (_("Failed to start session"));
@@ -228,15 +257,17 @@ start_session (void)
     g_free (session);
 }
 
-void user_treeview_selection_changed_cb (GtkTreeSelection *selection);
+void user_combobox_active_changed_cb (GtkComboBox *widget, LightDMGreeter *greeter);
 G_MODULE_EXPORT
 void
-user_treeview_selection_changed_cb (GtkTreeSelection *selection)
+user_combobox_active_changed_cb (GtkComboBox *widget, LightDMGreeter *greeter)
 {
     GtkTreeModel *model;
     GtkTreeIter iter;
 
-    if (gtk_tree_selection_get_selected (selection, &model, &iter))
+    model = gtk_combo_box_get_model (user_combo);
+
+    if (gtk_combo_box_get_active_iter (user_combo, &iter))
     {
         gchar *user;
 
@@ -417,14 +448,13 @@ user_added_cb (LightDMUserList *user_list, LightDMUser *user)
     GtkTreeModel *model;
     GtkTreeIter iter;
 
-    model = gtk_tree_view_get_model (user_view);
+    model = gtk_combo_box_get_model (user_combo);
 
     gtk_list_store_append (GTK_LIST_STORE (model), &iter);
     gtk_list_store_set (GTK_LIST_STORE (model), &iter,
                         0, lightdm_user_get_name (user),
                         1, lightdm_user_get_display_name (user),
                         2, lightdm_user_get_logged_in (user) ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL,
-                        /*3, pixbuf,*/
                         -1);
 }
 
@@ -433,7 +463,7 @@ get_user_iter (const gchar *username, GtkTreeIter *iter)
 {
     GtkTreeModel *model;
 
-    model = gtk_tree_view_get_model (user_view);
+    model = gtk_combo_box_get_model (user_combo);
   
     if (!gtk_tree_model_get_iter_first (model, iter))
         return FALSE;
@@ -461,12 +491,12 @@ user_changed_cb (LightDMUserList *user_list, LightDMUser *user)
     if (!get_user_iter (lightdm_user_get_name (user), &iter))
         return;
 
-    model = gtk_tree_view_get_model (user_view);
+    model = gtk_combo_box_get_model (user_combo);
+
     gtk_list_store_set (GTK_LIST_STORE (model), &iter,
                         0, lightdm_user_get_name (user),
                         1, lightdm_user_get_display_name (user),
                         2, lightdm_user_get_logged_in (user) ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL,
-                        /*3, pixbuf,*/
                         -1);
 }
 
@@ -479,7 +509,7 @@ user_removed_cb (LightDMUserList *user_list, LightDMUser *user)
     if (!get_user_iter (lightdm_user_get_name (user), &iter))
         return;
 
-    model = gtk_tree_view_get_model (user_view);  
+    model = gtk_combo_box_get_model (user_combo);
     gtk_list_store_remove (GTK_LIST_STORE (model), &iter);
 }
 
@@ -521,9 +551,15 @@ void
 a11y_contrast_cb (GtkWidget *widget)
 {
     if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (widget)))
+    {
         g_object_set (gtk_settings_get_default (), "gtk-theme-name", "HighContrastInverse", NULL);
+        g_object_set (gtk_settings_get_default (), "gtk-icon-theme-name", "HighContrastInverse", NULL);
+    }
     else
+    {
         g_object_set (gtk_settings_get_default (), "gtk-theme-name", default_theme_name, NULL);
+        g_object_set (gtk_settings_get_default (), "gtk-icon-theme-name", default_icon_theme_name, NULL);
+    }
 }
 
 static void
@@ -545,35 +581,17 @@ load_user_list ()
     g_signal_connect (lightdm_user_list_get_instance (), "user-changed", G_CALLBACK (user_changed_cb), NULL);
     g_signal_connect (lightdm_user_list_get_instance (), "user-removed", G_CALLBACK (user_removed_cb), NULL);
 
-    model = gtk_tree_view_get_model (user_view);
+    model = gtk_combo_box_get_model (user_combo);
     items = lightdm_user_list_get_users (lightdm_user_list_get_instance ());
     for (item = items; item; item = item->next)
     {
         LightDMUser *user = item->data;
-        const gchar *image;
-        GdkPixbuf *pixbuf = NULL;
-
-        image = lightdm_user_get_image (user);
-        if (image)
-            pixbuf = gdk_pixbuf_new_from_file_at_scale (image, 64, 64, TRUE, NULL);
-        if (!pixbuf)
-            pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
-                                               "stock_person",
-                                               64,
-                                               GTK_ICON_LOOKUP_USE_BUILTIN,
-                                               NULL);
-        /*if (!pixbuf)
-        {
-            pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 64, 64);
-            memset (gdk_pixbuf_get_pixels (pixbuf), 0, gdk_pixbuf_get_height (pixbuf) * gdk_pixbuf_get_rowstride (pixbuf) * gdk_pixbuf_get_n_channels (pixbuf));
-        }*/
 
         gtk_list_store_append (GTK_LIST_STORE (model), &iter);
         gtk_list_store_set (GTK_LIST_STORE (model), &iter,
                             0, lightdm_user_get_name (user),
                             1, lightdm_user_get_display_name (user),
                             2, lightdm_user_get_logged_in (user) ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL,
-                            3, pixbuf,
                             -1);
     }
     if (lightdm_greeter_get_has_guest_account_hint (greeter))
@@ -583,7 +601,6 @@ load_user_list ()
                             0, "*guest",
                             1, _("Guest Account"),
                             2, PANGO_WEIGHT_NORMAL,
-                            3, gtk_icon_theme_load_icon (gtk_icon_theme_get_default (), "stock_person", 64, 0, NULL),
                             -1);
     }
 
@@ -592,7 +609,6 @@ load_user_list ()
                         0, "*other",
                         1, _("Other..."),
                         2, PANGO_WEIGHT_NORMAL,
-                        3, gtk_icon_theme_load_icon (gtk_icon_theme_get_default (), "stock_person", 64, 0, NULL),
                         -1);
 
     last_user = g_key_file_get_value (state, "greeter", "last-user", NULL);
@@ -617,7 +633,7 @@ load_user_list ()
             g_free (name);
             if (matched)
             {
-                gtk_tree_selection_select_iter (gtk_tree_view_get_selection (user_view), &iter);
+                gtk_combo_box_set_active_iter (user_combo, &iter);
                 start_authentication (selected_user);
                 break;
             }
@@ -789,12 +805,22 @@ main (int argc, char **argv)
     value = g_key_file_get_value (config, "greeter", "theme-name", NULL);
     if (value)
     {
-        g_debug ("Using theme %s", value);
+        g_debug ("Using Gtk+ theme %s", value);
         g_object_set (gtk_settings_get_default (), "gtk-theme-name", value, NULL);
     }
     g_free (value);
     g_object_get (gtk_settings_get_default (), "gtk-theme-name", &default_theme_name, NULL);
-    g_debug ("Default theme is '%s'", default_theme_name);
+    g_debug ("Default Gtk+ theme is '%s'", default_theme_name);
+
+    value = g_key_file_get_value (config, "greeter", "icon-theme-name", NULL);
+    if (value)
+    {
+        g_debug ("Using icon theme %s", value);
+        g_object_set (gtk_settings_get_default (), "gtk-icon-theme-name", value, NULL);
+    }
+    g_free (value);
+    g_object_get (gtk_settings_get_default (), "gtk-icon-theme-name", &default_icon_theme_name, NULL);
+    g_debug ("Default theme is '%s'", default_icon_theme_name);
 
     value = g_key_file_get_value (config, "greeter", "font-name", NULL);
     if (value)
@@ -838,10 +864,37 @@ main (int argc, char **argv)
     prompt_label = GTK_LABEL (gtk_builder_get_object (builder, "prompt_label"));
     prompt_entry = GTK_ENTRY (gtk_builder_get_object (builder, "prompt_entry"));
     message_label = GTK_LABEL (gtk_builder_get_object (builder, "message_label"));
+    user_combo = GTK_COMBO_BOX (gtk_builder_get_object (builder, "user_combobox"));
     session_combo = GTK_COMBO_BOX (gtk_builder_get_object (builder, "session_combobox"));
     language_combo = GTK_COMBO_BOX (gtk_builder_get_object (builder, "language_combobox"));  
     panel_window = GTK_WINDOW (gtk_builder_get_object (builder, "panel_window"));
 
+    /* Load logo */
+    value = g_key_file_get_value (config, "greeter", "logo", NULL);
+    if (value)
+    {
+        gchar *path;
+        GtkImage *logo = GTK_IMAGE (gtk_builder_get_object (builder, "logo"));
+        GdkPixbuf *logo_pixbuf = NULL;
+        GError *error = NULL;
+
+        if (g_path_is_absolute (value))
+            path = g_strdup (value);
+        else
+            path = g_build_filename (GREETER_DATA_DIR, value, NULL);
+
+        g_debug ("Loading logo %s", path);
+        logo_pixbuf = gdk_pixbuf_new_from_file (path, &error);
+        if (logo_pixbuf)
+            gtk_image_set_from_pixbuf (logo, logo_pixbuf);
+        else
+           g_warning ("Failed to load logo: %s", error->message);
+        g_clear_error (&error);
+        g_object_unref (logo_pixbuf);
+        g_free (path);
+        g_free (value);
+    }
+
     gtk_label_set_text (GTK_LABEL (gtk_builder_get_object (builder, "hostname_label")), lightdm_get_hostname ());
 
     /* Glade can't handle custom menuitems, so set them up manually */
@@ -857,7 +910,7 @@ main (int argc, char **argv)
     hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
     gtk_widget_show (hbox);
     gtk_container_add (GTK_CONTAINER (menuitem), hbox);
-    image = gtk_image_new_from_icon_name ("accessibility", GTK_ICON_SIZE_MENU);
+    image = gtk_image_new_from_icon_name ("preferences-desktop-accessibility", GTK_ICON_SIZE_MENU);
     gtk_widget_show (image);
     gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, TRUE, 0);
 
@@ -928,22 +981,24 @@ main (int argc, char **argv)
         set_language (NULL);
     }
 
-    user_view = GTK_TREE_VIEW (gtk_builder_get_object (builder, "user_treeview"));
-    gtk_tree_view_insert_column_with_attributes (user_view, 0, "Face", gtk_cell_renderer_pixbuf_new(), "pixbuf", 3, NULL);
-    gtk_tree_view_insert_column_with_attributes (user_view, 1, "Name", gtk_cell_renderer_text_new(), "text", 1, "weight", 2, NULL);
+    renderer = gtk_cell_renderer_text_new();
+    gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (user_combo), renderer, TRUE);
+    gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (user_combo), renderer, "text", 1);
+    gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (user_combo), renderer, "weight", 2);
 
     if (lightdm_greeter_get_hide_users_hint (greeter))
         start_authentication ("*other");
     else
     {
         load_user_list ();
-        gtk_widget_show (GTK_WIDGET (user_view));
+        gtk_widget_show (GTK_WIDGET (user_combo));
     } 
 
     gtk_builder_connect_signals(builder, greeter);
 
     gtk_widget_show (GTK_WIDGET (login_window));
     center_window (login_window);
+    g_signal_connect (GTK_WIDGET (login_window), "size-allocate", G_CALLBACK (center_window), login_window);
 
     gtk_widget_show (GTK_WIDGET (panel_window));
     GtkAllocation allocation;
